@@ -15,9 +15,10 @@ from nose.tools import assert_raises
 from datetime import datetime, timedelta
 from werkzeug.wrappers import *
 from werkzeug.wsgi import LimitedStream
+from werkzeug.http import generate_etag
 from werkzeug.datastructures import MultiDict, ImmutableOrderedMultiDict, \
      ImmutableList, ImmutableTypeConversionDict
-from werkzeug.test import Client, create_environ
+from werkzeug.test import Client, create_environ, run_wsgi_app
 
 
 class RequestTestResponse(BaseResponse):
@@ -101,6 +102,47 @@ def test_base_request():
     assert response['form'] == MultiDict()
 
 
+def test_access_route():
+    """Check access route on the wrappers"""
+    req = Request.from_values(headers={
+        'X-Forwarded-For': '192.168.1.2, 192.168.1.1'
+    })
+    req.environ['REMOTE_ADDR'] = '192.168.1.3'
+    req.is_behind_proxy = True
+    assert req.access_route == ['192.168.1.2', '192.168.1.1']
+    assert req.remote_addr == '192.168.1.2'
+    req.is_behind_proxy = False
+    assert req.access_route == ['192.168.1.2', '192.168.1.1']
+    assert req.remote_addr == '192.168.1.3'
+
+    req = Request.from_values()
+    req.environ['REMOTE_ADDR'] = '192.168.1.3'
+    assert req.access_route == ['192.168.1.3']
+
+
+def test_url_request_descriptors():
+    """Basic URL request descriptors"""
+    req = Request.from_values('/bar?foo=baz', 'http://example.com/test')
+    assert req.path == u'/bar'
+    assert req.script_root == u'/test'
+    assert req.url == 'http://example.com/test/bar?foo=baz'
+    assert req.base_url == 'http://example.com/test/bar'
+    assert req.url_root == 'http://example.com/test/'
+    assert req.host_url == 'http://example.com/'
+    assert req.host == 'example.com'
+
+
+def test_authorization_mixin():
+    """Authorization mixin"""
+    request = Request.from_values(headers={
+        'Authorization': 'Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ=='
+    })
+    a = request.authorization
+    assert a.type == 'basic'
+    assert a.username == 'Aladdin'
+    assert a.password == 'open sesame'
+
+
 def test_base_response():
     """Base respone behavior"""
     # unicode
@@ -121,6 +163,48 @@ def test_base_response():
          '01-Jan-1970 00:00:00 GMT; Max-Age=60; Path=/blub')
     ]
 
+    # delete cookie
+    response = BaseResponse()
+    response.delete_cookie('foo')
+    assert response.headers.to_list() == [
+        ('Content-Type', 'text/plain; charset=utf-8'),
+        ('Set-Cookie', 'foo=; expires=Thu, 01-Jan-1970 00:00:00 GMT; Max-Age=0; Path=/')
+    ]
+
+    # close call forwarding
+    closed = []
+    class Iterable(object):
+        def next(self):
+            raise StopIteration()
+        def __iter__(self):
+            return self
+        def close(self):
+            closed.append(True)
+    response = BaseResponse(Iterable())
+    response.call_on_close(lambda: closed.append(True))
+    app_iter, status, headers = run_wsgi_app(response,
+                                             create_environ(),
+                                             buffered=True)
+    assert status == '200 OK'
+    assert ''.join(app_iter) == ''
+    assert len(closed) == 2
+
+
+def test_response_status_codes():
+    """Response status codes"""
+    response = BaseResponse()
+    response.status_code = 404
+    assert response.status == '404 NOT FOUND'
+    response.status = '200 OK'
+    assert response.status_code == 200
+    response.status = '999 WTF'
+    assert response.status_code == 999
+    response.status_code = 588
+    assert response.status_code == 588
+    assert response.status == '588 UNKNOWN'
+    response.status = 'wtf'
+    assert response.status_code == 0
+
 
 def test_type_forcing():
     """Response Type forcing"""
@@ -145,7 +229,7 @@ def test_type_forcing():
         assert response.content_type == 'text/html'
 
     # without env, no arbitrary conversion
-    assert_raises(TypeError, "SpecialResponse.force_type(wsgi_application)")
+    assert_raises(TypeError, SpecialResponse.force_type, wsgi_application)
 
 
 def test_accept_mixin():
@@ -168,6 +252,9 @@ def test_accept_mixin():
     assert request.accept_encodings == CharsetAccept([('gzip', 1), ('deflate', 1)])
     assert request.accept_languages == CharsetAccept([('en-us', 1), ('en', 0.5)])
 
+    request = Request({'HTTP_ACCEPT': ''})
+    assert request.accept_mimetypes == CharsetAccept()
+
 
 def test_etag_request_mixin():
     """ETag request-wrapper mixin"""
@@ -212,6 +299,12 @@ def test_user_agent_mixin():
         assert request.user_agent.platform == platform
         assert request.user_agent.version == version
         assert request.user_agent.language == lang
+        assert bool(request.user_agent)
+        assert request.user_agent.to_header() == ua
+        assert str(request.user_agent) == ua
+
+    request = Request({'HTTP_USER_AGENT': 'foo'})
+    assert not request.user_agent
 
 
 def test_etag_response_mixin():
@@ -241,6 +334,35 @@ def test_etag_response_mixin():
     assert not 'content-length' in resp.headers
 
 
+def test_etag_response_mixin_freezing():
+    """Freeze of the etag response mixin adds etag if mixed first"""
+    class WithFreeze(ETagResponseMixin, BaseResponse):
+        pass
+    class WithoutFreeze(BaseResponse, ETagResponseMixin):
+        pass
+
+    response = WithFreeze('Hello World')
+    response.freeze()
+    assert response.get_etag() == (generate_etag('Hello World'), False)
+    response = WithoutFreeze('Hello World')
+    response.freeze()
+    assert response.get_etag() == (None, None)
+    response = Response('Hello World')
+    response.freeze()
+    assert response.get_etag() == (None, None)
+
+
+def test_authenticate_mixin():
+    """Test the authenciate mixin of the response"""
+    resp = Response()
+    resp.www_authenticate.type = 'basic'
+    resp.www_authenticate.realm = 'Testing'
+    assert resp.headers['WWW-Authenticate'] == 'Basic realm="Testing"'
+    resp.www_authenticate.realm = None
+    resp.www_authenticate.type = None
+    assert 'WWW-Authenticate' not in resp.headers
+
+
 def test_response_stream_mixin():
     """Response stream response-wrapper mixin"""
     response = Response()
@@ -476,13 +598,18 @@ def test_storage_classes():
     class MyRequest(Request):
         dict_storage_class = dict
         list_storage_class = list
-    req = MyRequest.from_values(headers={
+        parameter_storage_class = dict
+    req = MyRequest.from_values('/?foo=baz', headers={
         'Cookie':   'foo=bar'
     })
     assert type(req.cookies) is dict
     assert req.cookies == {'foo': 'bar'}
     assert type(req.access_route) is list
 
+    assert type(req.args) is dict
+    assert type(req.values) is CombinedMultiDict
+    assert req.values['foo'] == 'baz'
+
     req = Request.from_values(headers={
         'Cookie':   'foo=bar'
     })
@@ -493,3 +620,10 @@ def test_storage_classes():
     MyRequest.list_storage_class = tuple
     req = MyRequest.from_values()
     assert type(req.access_route) is tuple
+
+
+def test_response_headers_passthrough():
+    """If headers are a Headers object they will be stored on the response"""
+    headers = Headers()
+    resp = Response(headers=headers)
+    assert resp.headers is headers
